"""The DataEcho class provides a way to echo data from one object to
another."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import time
from typing import Callable, Any

import numpy as np
from icecream import ic
from vistutils.fields import MutableDescriptor, unParseArgs
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

from vistside.core import ArrayField, RightNowField


class _Im:
  """Retrieves the imaginary part of the series."""

  def __get__(self, instance: DataRoll, owner: type) -> np.ndarray:
    """Returns the imaginary part of the last value."""
    return instance.toArray().imag


class _Re:
  """Retrieves the real part of the series."""

  def __get__(self, instance: DataRoll, owner: type) -> np.ndarray:
    """Returns the real part of the last value."""
    return instance.toArray().real


class DataRoll:
  """The DataEcho class provides a way to echo data from one object to
  another."""

  __callback_functions__ = None
  __zero_index__ = None

  imag = _Im()
  real = _Re()
  rightNow = RightNowField()
  array = ArrayField(16)

  def __init__(self, *args, **kwargs) -> None:
    self.__callback_functions__ = []
    self.__zero_index__ = 0

  def appendCallback(self, callMeMaybe: Callable) -> Callable:
    """Notify the DataEcho object when a value is appended."""
    if not callable(callMeMaybe):
      e = typeMsg('callMeMaybe', callMeMaybe, Callable)
      raise TypeError(e)
    self.__callback_functions__.append(callMeMaybe)
    return callMeMaybe

  def CALL(self, callMeMaybe: Callable) -> Callable:
    """Call the callback functions."""
    return self.appendCallback(callMeMaybe)

  def notifyCallbacks(self, val: complex) -> None:
    """Notify the callback functions."""
    for func in self.__callback_functions__:
      func()

  def append(self, value: float = None) -> None:
    """Append a value to the DataEcho object."""
    if not value == value:
      return
    if isinstance(value, int):
      value = float(value) + 0j
    if isinstance(value, complex):
      value = value.imag
    self.array[self.__zero_index__] = self.rightNow + value * 1j
    entry = None
    try:
      entry = self.array[self.__zero_index__]
      self.notifyCallbacks(entry.item())
    except ValueError as valueError:
      ic(entry)
      ic(self.__zero_index__)
      raise valueError
    self.__zero_index__ = (self.__zero_index__ + 1) % len(self.array)

  def __str__(self, ) -> str:
    line = []
    lines = []
    for val in self.toArray():
      this = '(%.3f + %.3fI)' % (val.real, val.imag)
      if sum([len(x) for x in line]) + len(this) + 2 * len(line) + 1 > 77:
        lines.append(', '.join(line))
        line = []
      line.append(this)
    if line:
      lines.append(', '.join(line))
    return '\n'.join(lines)

  def __len__(self, ) -> int:
    """Returns the length of the array  ."""
    return self.array.shape[0]

  def toArray(self, ) -> np.ndarray:
    """Returns the array."""
    arr = np.roll(self.array, -self.__zero_index__)
    return arr[~np.isnan(arr)]

  @classmethod
  def getDefault(cls, *args, **kwargs) -> DataRoll:
    """Returns the default value for the field."""
    dataRoll = DataRoll()
    return dataRoll.apply((args, kwargs))

  def apply(self, value: Any) -> DataRoll:
    """Applies the value to the field."""
    args, kwargs = unParseArgs(value)
    return self


class DataRollField(MutableDescriptor):
  """Wraps the DataRoll in a mutable descriptor class"""

  def __init__(self, *args, **kwargs) -> None:
    MutableDescriptor.__init__(self, DataRoll, *args, **kwargs)

  def __get__(self, *args, **kwargs) -> DataRoll:
    """Returns the value of the field."""
    return super().__get__(*args, **kwargs)



"""HelpMenu provides the help menu of the main application window"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from vistside.windows.actions import AbstractMenu


class HelpMenu(AbstractMenu):
  """HelpMenu provides the help menu of the main application window"""

  def setupActions(self, ) -> None:
    """setupActions creates the actions for the menu."""
    self.parent().aboutAction = self.addAction('About', )
    self.parent().helpAction = self.addAction('Help', )
    self.parent().aboutQtAction = self.addAction('About Qt', )



"""BaseWidget provides the lowest level of abstraction for a QWidget. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any, Self

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QWidget
from vistutils.fields import unParseArgs

from vistside.core import TextPen, EmptyPen, EmptyBrush, parseParent


class BaseWidget(QWidget):
  """BaseWidget provides the lowest level of abstraction for a QWidget. """

  hover = Signal()
  leave = Signal()
  leftClick = Signal()
  rightClick = Signal()
  middleClick = Signal()
  forwardClick = Signal()
  backwardClick = Signal()
  doubleClick = Signal()
  leftPressHold = Signal()
  rightPressHold = Signal()

  textPen = TextPen()
  emptyPen = EmptyPen()
  emptyBrush = EmptyBrush()

  def __init__(self, *args, **kwargs) -> None:
    parent = parseParent(*args)
    QWidget.__init__(self, parent)

  @classmethod
  def getDefault(cls, *args, **kwargs) -> Self:
    """Creates a default instance"""
    return BaseWidget()

  def apply(self, value: Any) -> Self:
    """Applies the value to the field."""
    args, kwargs = unParseArgs(value)
    return self



"""FillWidget provides a widget with a fill color"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QPaintEvent, QPainter

from vistside.core import BrushField
from vistside.widgets import BaseWidget


class FillWidget(BaseWidget):
  """FillWidget provides a widget with a fill color"""

  fillBrush = BrushField()

  def __init__(self, *args, **kwargs) -> None:
    BaseWidget.__init__(self, *args, **kwargs)
    self.fillBrush = 144, 0, 255, 255

  def paintEvent(self, event: QPaintEvent) -> None:
    """Paints the widget."""
    painter = QPainter()
    painter.begin(self)
    painter.setRenderHint(painter.Antialiasing)
    painter.setPen(self.emptyPen)
    painter.setBrush(self.fillBrush)
    painter.drawRect(painter.viewport())
    painter.end()



"""AbstractMenu provides a base class for the menus of the main
application window. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu
from vistutils.parse import maybe
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

from vistside.core import KeyboardShortcuts, parseParent
from vistside.core.icons import getIcon


class AbstractMenu(QMenu):
  """AbstractMenu provides a base class for the menus of the main
  application window. """

  @staticmethod
  def _parseTitle(*args, **kwargs) -> str:
    """Parses argument to find title"""
    titleKeys = stringList("""title, menuTitle, menu, name, titleName""")
    titleKwarg = None
    for key in titleKeys:
      if key in kwargs:
        val = kwargs.get(key)
        if isinstance(val, str):
          titleKwarg = val
          break
    titleArg = [*[arg for arg in args if isinstance(arg, str)], None][0]
    title = maybe(titleKwarg, titleArg, )
    if isinstance(title, str):
      return title
    e = typeMsg('title', title, str)
    raise TypeError(e)

  def __init__(self, *args, **kwargs) -> None:
    parent = parseParent(*args)
    title = self._parseTitle(*args, **kwargs)
    QMenu.__init__(self, title, parent)
    self.menuAction().setIconText('title')

  def addAction(self, *args, ) -> QAction:
    """addAction creates a new action and adds it to the menu. """
    name = args[0]
    try:
      icon = getIcon(name)
    except FileNotFoundError as e:
      icon = getIcon('risitas')
    keyboard = KeyboardShortcuts[name.replace(' ', '_').lower()]
    action = QAction(icon, name, self)
    action.setShortcut(keyboard)
    action.setShortcutVisibleInContextMenu(True)
    QMenu.addAction(self, action)
    return action



"""MainStatusBar provides the status bar for the main application window. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from PySide6.QtWidgets import QStatusBar, QMainWindow, QLabel
from icecream import ic
from vistutils.fields import unParseArgs

from vistside.core import parseParent
from vistside.widgets import ClockField


class MainStatusBar(QStatusBar):
  """MainStatusBar provides the status bar for the main application
  window. """

  clock = ClockField()

  def __init__(self, *args, **kwargs) -> None:
    parent = parseParent(*args)
    QStatusBar.__init__(self, parent)
    self.setStyleSheet("""
    QStatusBar {
        margin: 0px;
        padding: 0px;
        background: rgba(255, 255, 255, 0.5);
        color: black;
        font-family: "Courier", sans-serif; 
        font-size: 14px;

    }""")

  def show(self) -> None:
    """Hook to print to stdout"""
    ic(self)
    QStatusBar.show(self, )

  @classmethod
  def getDefault(cls, *args, **kwargs) -> MainStatusBar:
    """Returns the default MainStatusBar"""
    mainWindow = parseParent(*args)
    statusBar = cls(mainWindow)
    statusBar.clock.initUI()
    statusBar.clock.connectActions()
    statusBar.addPermanentWidget(statusBar.clock)
    statusBar.apply((args, kwargs))
    return statusBar

  def apply(self, value: Any) -> MainStatusBar:
    """Applies the value to the field."""
    args, kwargs = unParseArgs(value)
    for arg in args:
      if isinstance(arg, str):
        self.showMessage(arg)
    return self



"""TextPenField provides a descriptor for the QPen used to print text."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from PySide6.QtGui import QColor, QPen
from vistutils.fields import StaticField
from vistutils.parse import maybe
from vistutils.waitaminute import typeMsg

from vistside.core import parseColor, SolidLine


class TextPen(StaticField):
  """TextPenField provides a descriptor for the QPen used to print text."""

  __font_color__ = None
  __fallback_color__ = QColor(0, 0, 0, 255)

  def __init__(self, *args, **kwargs) -> None:
    col = maybe(parseColor(*args, **kwargs), self.__fallback_color__)
    if isinstance(col, QColor):
      self.__font_color__ = col
    else:
      e = typeMsg('col', col, QColor)
      raise TypeError(e)
    pen = QPen()
    pen.setStyle(SolidLine)
    pen.setColor(self.__font_color__)
    pen.setWidth(1)
    StaticField.__init__(self, pen)

  def __set__(self, instance: object, value: Any) -> None:
    if isinstance(value, QColor):
      self.__font_color__ = value
      pen = self.__field_value__
      if isinstance(pen, QPen):
        pen.setColor(self.__font_color__)
      else:
        e = typeMsg('pen', pen, QPen)
        raise TypeError(e)
      pen.setColor(self.__font_color__)
      setattr(self, '__field_value__', pen)
    else:
      try:
        self.__set__(instance, parseColor(value))
      except Exception as exception:
        e = typeMsg('value', value, QColor)
        raise TypeError(e) from exception



"""This file provides standard keyboard shortcuts. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QKeySequence


class KeyboardShortcuts:
  """This class provides standard keyboard shortcuts. """

  __output_mode__ = str

  # File menu
  new = "Ctrl+N"
  open = "Ctrl+O"
  save = "Ctrl+S"
  save_as = "Ctrl+Shift+S"
  close = "Ctrl+W"
  quit = "Ctrl+Q"

  # Edit menu
  undo = "Ctrl+Z"
  redo = "Ctrl+Y"
  cut = "Ctrl+X"
  copy = "Ctrl+C"
  paste = "Ctrl+V"
  select_all = "Ctrl+A"

  # Help menu
  debug01 = 'F1'
  debug02 = 'F2'
  debug03 = 'F3'
  debug04 = 'F4'
  debug05 = 'F5'
  debug06 = 'F6'
  debug07 = 'F7'
  debug08 = 'F8'
  debug09 = 'F9'
  debug10 = 'F10'

  about = 'F11'
  about_qt = 'F12'

  @classmethod
  def getQKey(cls, name: str, **kwargs) -> QKeySequence:
    """getQKey returns the QKeySequence for a keyboard shortcut. """
    shortcutString = getattr(cls, name, '')
    out = QKeySequence.fromString(shortcutString)
    if out:
      return out
    if kwargs.get('_recursion', False):
      raise RecursionError
    return cls.getQKey(name.lower(), _recursion=True)

  @classmethod
  def __class_getitem__(cls, *args, **kwargs) -> QKeySequence:
    """__call__ returns the QKeySequence for a keyboard shortcut. """
    return cls.getQKey(*args, **kwargs)



"""The 'paint' modules provides factory functions for modular painting
operations. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._abstract_paint import AbstractPaint
from ._border_rect import BorderRect
from ._fill_rect import FillRect
from ._data_paint import DataPaint
from ._text_rect import TextRect



"""BorderRect draws an outline on the update rectangle, without fill."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QPaintEvent, QPainter

from ezros.gui.factories import emptyBrush, parsePen
from ezros.gui.paint import AbstractPaint
from ezros.gui.shortnames import Black


# from _dep.morevistutils import Wait


class BorderRect(AbstractPaint):
  """BorderRect draws an outline on the update rectangle, without fill."""

  # pen = Wait(parsePen, Black, 2)

  def paintOp(self, event: QPaintEvent, painter: QPainter) -> None:
    """Applies the paint operation"""
    painter.setPen(self.pen)
    painter.setBrush(emptyBrush())
    painter.drawRect(event.rect())

  def __init__(self, *args, **kwargs) -> None:
    self.pen = parsePen(*args, Black, 2, **kwargs)



"""LayoutWindow subclasses the BaseWindow and provides the visual layouts."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from vistside.core import parseParent
from vistside.widgets import LabelField
from vistside.windows import BaseWindow


class LayoutWindow(BaseWindow):
  """LayoutWindow subclasses the BaseWindow and provides the visual
  layouts."""

  welcomeBanner = LabelField('Yolo!', 36)

  def __init__(self, *args, **kwargs) -> None:
    BaseWindow.__init__(self, *args, **kwargs)

  def show(self) -> None:
    """Shows the window. The base window class already provides the base
    widget and base layout. This method should add widgets to the base
    layout and then call the parent show method."""
    self.baseLayout.addWidget(self.welcomeBanner)
    BaseWindow.show(self)



"""ArrayField class for storing and manipulating arrays of data."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import time
from typing import Any

import numpy as np
from vistutils.fields import Wait, MutableDescriptor, IntField
from vistutils.text import stringList, monoSpace
from vistutils.waitaminute import typeMsg


class ArrayField(MutableDescriptor):
  """ArrayField class for storing and manipulating arrays of data."""

  __fallback_num_data__ = 128

  numData = IntField()

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the field."""
    numKeys = stringList("""numData, numDataPoints, numDataPoints, n, num""")
    for key in numKeys:
      if key in kwargs:
        val = kwargs.get(key)
        if isinstance(val, int):
          self.numData = val
          break
        e = typeMsg(key, val, int)
        raise TypeError(e)
    else:
      for arg in args:
        if isinstance(arg, int):
          self.numData = arg
          break
      else:
        self.numData = self.__fallback_num_data__
    MutableDescriptor.__init__(self, np.ndarray, *args, **kwargs)
    self._setFieldType(np.ndarray)

  def _instantiate(self, instance: object) -> None:
    """Instantiates the field."""
    pvtName = self._getPrivateName()
    array = np.full((self.numData,), np.nan, dtype=complex)
    setattr(instance, self._getPrivateName(), array)

  def __get__(self, instance: object, owner: type, **kwargs) -> Any:
    """Returns the value of the field."""
    if instance is None:
      return self
    pvtName = self._getPrivateName()
    if getattr(instance, pvtName, None) is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self._instantiate(instance)
      return self.__get__(instance, owner, _recursion=True)
    return getattr(instance, self._getPrivateName())

  def __set__(self, instance: object, value: Any) -> None:
    """Sets the value of the field."""
    e = """This subclass of MutableDescriptor does not implement a setter 
    operation. """
    raise TypeError(monoSpace(e))

  def __delete__(self, instance: object, ) -> None:
    """Deletes the value of the field."""
    e = """Instances of MutableDescriptor cannot be deleted."""
    raise TypeError(monoSpace(e))



"""FilesMenu subclasses QMenu and provides a menu for file actions. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from vistside.windows.actions import AbstractMenu

if TYPE_CHECKING:
  pass


class FilesMenu(AbstractMenu):
  """FilesMenu subclasses QMenu and provides a menu for file actions. """

  def setupActions(self, ) -> None:
    """setupActions creates the actions for the menu."""
    self.parent().newAction = self.addAction('New', )
    self.parent().openAction = self.addAction('Open', )
    self.parent().saveAction = self.addAction('Save', )
    self.parent().saveAsAction = self.addAction('Save As', )
    self.parent().exitAction = self.addAction('Exit', )



"""EmptyBrush provides a descriptor returning the empty brush"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Qt
from PySide6.QtGui import QColor, QBrush
from vistutils.fields import StaticField


class EmptyBrush(StaticField):
  """EmptyBrush provides a descriptor returning the empty brush"""

  def __init__(self, ) -> None:
    brush = QBrush()
    brush.setColor(QColor(0, 0, 0, 0, ))
    brush.setStyle(Qt.BrushStyle.NoBrush)
    StaticField.__init__(self, brush)



"""GridLayout wraps the QGridLayout."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QGridLayout

from vistside.core import parseParent


class GridLayout(QGridLayout):
  """GridLayout wraps the QGridLayout."""

  __field_name__ = None
  __field_owner__ = None

  def __init__(self, *args, **kwargs) -> None:
    parent = parseParent(*args)
    QGridLayout.__init__(self, parent)
    self.setContentsMargins(2, 2, 2, 2, )
    self.setSpacing(2)

  def __set_name__(self, owner: type, name: str) -> None:
    self.__field_owner__ = owner
    self.__field_name__ = name



"""JKFlipFlop implementing a toggle button with separate clickable areas
for activation and deactivation."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any
from PySide6.QtWidgets import QWidget, QPushButton, QVBoxLayout
from PySide6.QtCore import Qt, QRect, Signal
from PySide6.QtGui import QPainter, QBrush, QColor

from vistside.widgets import BaseWidget


class JKFlipFlop(BaseWidget):
  """
  A toggle button widget that acts like a JK flip-flop,
  with separate buttons for activation and deactivation.
  """

  stateChanged = Signal()
  turnOn = Signal()
  turnOff = Signal()

  def __init__(self, *args, **kwargs) -> None:
    BaseWidget.__init__(self, *args, **kwargs)
    self.activateButton = QPushButton("Activate", self)
    self.deactivateButton = QPushButton("Deactivate", self)
    self.activateButton.clicked.connect(self.activate)
    self.activateButton.clicked.connect(self.turnOn.emit)
    self.activateButton.clicked.connect(self.stateChanged.emit)
    self.deactivateButton.clicked.connect(self.turnOff.emit)
    self.deactivateButton.clicked.connect(self.deactivate)
    self.deactivateButton.clicked.connect(self.stateChanged.emit)
    self.state: bool = False
    self.initUI()

  def initUI(self) -> None:
    """Initializes the user interface."""
    layout = QVBoxLayout()
    layout.addWidget(self.activateButton)
    layout.addWidget(self.deactivateButton)
    self.setLayout(layout)
    self.deactivate()

  def activate(self) -> None:
    """Activates the toggle."""
    self.state = True
    self.activateButton.setDisabled(True)
    self.deactivateButton.setDisabled(False)
    self.update()

  def deactivate(self) -> None:
    """Deactivates the toggle."""
    self.state = False
    self.activateButton.setDisabled(False)
    self.deactivateButton.setDisabled(True)
    self.update()

  def paintEvent(self, event: Any) -> None:
    """Custom paint event to visually represent the toggle state."""
    painter = QPainter(self)
    if self.state:
      painter.setBrush(QBrush(QColor(0, 255, 0)))  # Green for active
    else:
      painter.setBrush(QBrush(QColor(255, 0, 0)))  # Red for inactive
    painter.drawRoundedRect(painter.viewport(), 8, 8)



"""QuickField provides a descriptor class requiring the owner to
explicitly decorate  the getter function and the creator function."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Callable


class DescriptorException(Exception):
  """DescriptorException should be raised by getters that require the
  creator function to run. They try-except block in the descriptor
  implementation will catch this exception and run the creator function. """

  def __init__(self, *args, **kwargs) -> None:
    allArgs = [*args, None, None, None, None][:4]
    ownerClass, instanceName, fieldName, descriptorClass = allArgs
    e = """Instance '%s' of type: '%s' has not yet any support for the 
    field at name: '%s', but the getter-function of the '%s' descriptor 
    class was still invoked. The creator function will now run to create 
    the missing value. The getter is then called again, but  with a 
    recursion flag that raises a recursion error if again, the name is not 
    ready. """ % (instanceName, ownerClass, fieldName, descriptorClass)
    Exception.__init__(self, e)


class QuickField:
  """QuickField provides a descriptor class requiring the owner to
  explicitly decorate  the getter function and the creator function. """

  __field_name__ = None
  __field_owner__ = None
  __instance_creator__ = None
  __instance_getter__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner. """
    self.__field_name__ = name
    self.__field_owner__ = owner

  def _setCreator(self, creator: Callable) -> Callable:
    """Set the creator function for the field. """
    self.__instance_creator__ = creator
    return creator

  def _getCreator(self) -> Callable:
    """Get the creator function for the field. """
    return self.__instance_creator__

  def _setGetter(self, getter: Callable) -> Callable:
    """Set the getter function for the field. """
    self.__instance_getter__ = getter
    return getter

  def _getGetter(self, ) -> Callable:
    """Get the getter function for the field. """
    return self.__instance_getter__

  def __get__(self, instance: object, owner: type, **kwargs) -> object:
    """Get the field from the instance. """
    getter = self._getGetter()
    if getter is None:
      e = f"""Getter not defined for field {self.__field_name__}"""
      raise AttributeError(e)
    try:
      return getter(instance)
    except DescriptorException as d:
      creator = self._getCreator()
      if kwargs.get('_recursion', False):
        raise RecursionError(d)
      if creator is None:
        e = f"""Creator not defined for field {self.__field_name__}"""
        raise AttributeError(e)
      if callable(creator):
        creator(instance)
      return getter(instance)



"""The quickFactories module contains a collection of factory functions for
creating QBrush, QPen and QFont instances. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Optional

from PySide6.QtGui import QColor
from vistutils.waitaminute import typeMsg

colorNames = {
  'red'         : QColor(255, 0, 0, 255),
  'green'       : QColor(0, 128, 0, 255),
  'blue'        : QColor(0, 0, 255, 255),
  'yellow'      : QColor(255, 255, 0, 255),
  'cyan'        : QColor(0, 255, 255, 255),
  'magenta'     : QColor(255, 0, 255, 255),
  'white'       : QColor(255, 255, 255, 255),
  'black'       : QColor(0, 0, 0, 255),
  'gray'        : QColor(128, 128, 128, 255),
  'lightgray'   : QColor(211, 211, 211, 255),
  'darkgray'    : QColor(169, 169, 169, 255),
  'lightred'    : QColor(255, 182, 193, 255),
  'darkred'     : QColor(139, 0, 0, 255),
  'lightgreen'  : QColor(144, 238, 144, 255),
  'darkgreen'   : QColor(0, 100, 0, 255),
  'lightblue'   : QColor(173, 216, 230, 255),
  'darkblue'    : QColor(0, 0, 139, 255),
  'lightyellow' : QColor(255, 255, 224, 255),
  'darkyellow'  : QColor(204, 204, 0, 255),
  'lightcyan'   : QColor(224, 255, 255, 255),
  'darkcyan'    : QColor(0, 139, 139, 255),
  'lightmagenta': QColor(255, 182, 193, 255),
  'darkmagenta' : QColor(139, 0, 139, 255),
}


def _parseIntColor(*color) -> Optional[QColor]:
  """Parses color from integer values if possible"""
  intArgs = [*[arg for arg in color if isinstance(arg, int)], ]
  if len(intArgs) < 3:
    return
  return QColor(*[*[arg for arg in color if isinstance(arg, int)], 255][:4])


def _parseQColor(*color) -> Optional[QColor]:
  """Parses color from QColor if possible"""
  for arg in color:
    if isinstance(arg, QColor):
      return arg


def _parseNamedColors(*color) -> Optional[QColor]:
  """Parses color from named color if possible"""
  for arg in color:
    if isinstance(arg, str):
      return QColor(arg)


def _argColor(*args) -> Optional[QColor]:
  """Parses color from arguments if possible"""
  for parser in [
    _parseQColor,
    _parseIntColor,
    _parseNamedColors,
  ]:
    result = parser(*args)
    if result is not None:
      if isinstance(result, QColor):
        return result
      e = typeMsg('result', result, QColor)
      raise TypeError(e)
    for arg in args:
      if isinstance(arg, (tuple, list)):
        result = _argColor(*arg)
        if result is not None:
          if isinstance(result, QColor):
            return result
          e = typeMsg('result', result, QColor)
          raise TypeError(e)


def _kwargColor(**kwargs) -> Optional[QColor]:
  """Finds color from keyword arguments"""
  r, g, b, a, col = None, None, None, 255, None
  for (key, val) in kwargs.items():
    if key in ['r', 'red']:
      r = val
    if key in ['g', 'green']:
      g = val
    if key in ['b', 'blue']:
      b = val
    if key in ['a', 'alpha']:
      a = val
    if key in ['color']:
      return _argColor(val)
  if all([isinstance(c, int) for c in [r, g, b, a]]):
    return QColor(r, g, b, a)


def parseColor(*args, **kwargs) -> QColor:
  """Parses color from arguments and keyword arguments"""
  colorKwarg = _kwargColor(**kwargs)
  if colorKwarg is not None:
    if isinstance(colorKwarg, QColor):
      return colorKwarg
    e = typeMsg('colorKwarg', colorKwarg, QColor)
    raise TypeError(e)
  colorArg = _argColor(*args)
  if colorArg is not None:
    if isinstance(colorArg, QColor):
      return colorArg
    e = typeMsg('colorArg', colorArg, QColor)
    raise TypeError(e)
  if kwargs.get('strict', False):
    argsE = """Unable to parse given arguments to instance of QColor! <br>
    Received args: %s <br>""" % '<br><tr>'.join([str(arg) for arg in args])
    kwargsE = """Received kwargs: %s""" % '<br><tr>'.join([
      '%s: %s' % (key, val) for (key, val) in kwargs.items()
    ])
    if not args and not kwargs:
      e = """No arguments or keyword arguments received!"""
      raise ValueError(e)
    if args and not kwargs:
      raise ValueError(argsE)
    if not args and kwargs:
      raise ValueError(kwargsE)
    raise ValueError(argsE + kwargsE)



"""PenField provides a descriptor class for instances of QPen."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from PySide6.QtGui import QPen
from vistutils.fields import CoreDescriptor


class PenField(CoreDescriptor):
  """The PenField class provides a descriptor for instances of QPen."""

  def _instantiate(self, instance: object) -> None:
    """Instantiates the field."""
    pen = QPen(*self._getArgs(), **self._getKwargs())
    setattr(instance, self._getPrivateName(), pen)

  def __set__(self, instance: object, value: Any) -> None:
    """Sets the field."""

    if not isinstance(value, QPen):
      e = """Value must be an instance of QPen!"""
      raise TypeError(e)
    setattr(instance, self._getPrivateName(), value)



"""MouseAware subclasses BaseWidget and provides methods relating to mouse
events"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Signal, Slot
from PySide6.QtGui import QMouseEvent

from vistside.widgets import BaseWidget
from vistutils.fields import Flag, IntField
from vistside.core import MouseBtn, LeftBtn, RightBtn, \
  MiddleBtn, NextBtn, BackBtn, NoBtn


class MouseAware(BaseWidget):
  """The MouseAware class provides methods relating to mouse events"""

  buttonClick = Signal(MouseBtn)
  doubleClick = Signal(MouseBtn)
  mouseEnter = Signal()
  mouseLeave = Signal()
  underMouse = Flag(False)
  pressedButton = ButtonField(NoBtn)
  DoubleButton = ButtonField(NoBtn)
  mouseLocalX = IntField(0)
  mouseLocalY = IntField(0)

  leftSingleClick = Signal()
  middleSingleClick = Signal()
  rightSingleClick = Signal()
  nextSingleClick = Signal()
  backSingleClick = Signal()
  leftDoubleClick = Signal()
  rightDoubleClick = Signal()
  middleDoubleClick = Signal()
  nextDoubleClick = Signal()
  backDoubleClick = Signal()

  @Slot(MouseBtn)
  def emitSingleClick(self, button: MouseBtn) -> None:
    """Filter click signals."""
    if button == LeftBtn:
      self.leftSingleClick.emit()
    if button == RightBtn:
      self.rightSingleClick.emit()
    if button == MiddleBtn:
      self.middleSingleClick.emit()
    if button == BackBtn:
      self.backSingleClick.emit()
    if button == NextBtn:
      self.nextSingleClick.emit()

  @Slot(MouseBtn)
  def emitDoubleClick(self, button: MouseBtn) -> None:
    """Filter double click signals."""
    if button == LeftBtn:
      self.leftDoubleClick.emit()
    if button == RightBtn:
      self.rightDoubleClick.emit()
    if button == MiddleBtn:
      self.middleDoubleClick.emit()
    if button == BackBtn:
      self.backDoubleClick.emit()
    if button == NextBtn:
      self.nextDoubleClick.emit()

  def mousePressEvent(self, event: QMouseEvent) -> None:
    """Mouse press event handler."""
    if not self.underMouse:
      return
    self.pressedButton = event.button()

  def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
    """Mouse double click event handler."""
    if not self.underMouse:
      return
    if event.button() == self.pressedButton:
      self.DoubleButton = event.button()
    else:
      self.DoubleButton = NoBtn

  def mouseReleaseEvent(self, event: QMouseEvent) -> None:
    """Mouse release event handler."""
    if not self.underMouse or event.button() == NoBtn:
      return
    if event.button() == self.DoubleButton:
      self.DoubleButton = NoBtn
      self.doubleClick.emit(event.button())
    if event.button() == self.pressedButton:
      self.buttonClick.emit(event.button())
    self.pressedButton = NoBtn

  def enterEvent(self, event: QMouseEvent) -> None:
    """Mouse enter event handler."""
    self.underMouse = True
    self.mouseEnter.emit()
    mousePos = event.pos()
    self.mouseLocalX = mousePos.x()
    self.mouseLocalY = mousePos.y()

  def leaveEvent(self, event: QMouseEvent) -> None:
    """Mouse leave event handler."""
    self.underMouse = False
    self.mouseLeave.emit()
    mousePos = event.pos()
    self.mouseLocalX = -mousePos.x()
    self.mouseLocalY = -mousePos.y()

  def mouseMoveEvent(self, event) -> None:
    """Mouse move event handler."""
    mousePos = event.pos()
    self.mouseLocalX = mousePos.x()
    self.mouseLocalY = mousePos.y()



"""BrushField provides a descriptor class for instances of QBrush."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from PySide6.QtGui import QBrush
from vistutils.fields import CoreDescriptor, Wait, unParseArgs

from vistside.core import parseBrush


class Brush(QBrush):
  """Subclass of QBrush with a custom constructor enabling interaction
  with descriptors."""

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the Brush."""
    QBrush.__init__(self, *args, **kwargs)

  @classmethod
  def getDefault(cls, *args, **kwargs) -> Brush:
    """Returns the default value for the field."""
    defVal = cls()
    defVal.apply((args, kwargs))
    return defVal

  def apply(self, value: Any) -> Brush:
    """Applies the value to the field."""
    args, kwargs = unParseArgs(value)
    testBrush = parseBrush(*args, **kwargs)
    self.setStyle(testBrush.style())
    self.setColor(testBrush.color())

    return self


class BrushField(Wait):
  """The BrushField class provides a descriptor for instances of QBrush."""

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the BrushField."""
    Wait.__init__(self, Brush, *args, **kwargs)



"""The FillRect factory creates a function applying solid paint to the
underlying device"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Self

from PySide6.QtGui import QPainter, QPaintEvent

from ezros.gui.factories import emptyPen, parseBrush, parseColor
from ezros.gui.paint import AbstractPaint
from ezros.gui.shortnames import Silver
from ezros.rosutils import Wait


# from _dep.morevistutils import Wait


class FillRect(AbstractPaint):
  """The FillRect factory creates a function applying solid paint to the
  underlying device"""

  brush = Wait(parseBrush, Silver, )

  def __init__(self, *args, **kwargs) -> None:
    pass
    # color = parseColor(*args, **kwargs)

  def paintOp(self, event: QPaintEvent, painter: QPainter) -> None:
    """Fills the rect with the given brush"""
    painter.setBrush(self.brush)
    painter.setPen(emptyPen())
    painter.drawRect(event.rect())

  def update(self, *args, **kwargs) -> Self:
    """Updates the paint settings. """
    if args:
      color = parseColor(*args, **kwargs)
      self.brush = parseBrush(color)
    return self



"""The brushFactory function creates images of QBrush."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Qt
from PySide6.QtGui import QBrush, QColor
from vistutils.text import stringList
from vistutils.parse import maybe

from vistside.core import parseColor


def _createBrush(*args, **kwargs) -> QBrush:
  """Creates a QBrush instance."""
  errors = []
  colorDefault = QColor(0, 0, 0, 255)
  styleDefault = Qt.BrushStyle.SolidPattern
  colorArg, styleArg = None, None
  colorKeys = stringList("""color, col, brushColor, rgb, RGBA""")
  styleKeys = stringList("""style, brushStyle""")
  for arg in args:
    if isinstance(arg, QColor) and colorArg is None:
      colorArg = arg
    elif isinstance(arg, Qt.BrushStyle) and styleArg is None:
      styleArg = arg
  colorKwarg, styleKwarg = None, None
  for key in colorKeys:
    if key in kwargs and colorKwarg is None:
      val = kwargs[key]
      try:
        colorKwarg = parseColor(val)
      except ValueError as valueError:
        errors.append(valueError)
  for key in styleKeys:
    if key in kwargs and styleKwarg is None:
      val = kwargs[key]
      if isinstance(val, Qt.BrushStyle):
        styleKwarg = val
  color = maybe(colorKwarg, colorArg, colorDefault)
  style = maybe(styleKwarg, styleArg, styleDefault)
  brush = QBrush()
  brush.setColor(color)
  brush.setStyle(style)
  return brush


def parseBrush(*args, **kwargs) -> QBrush:
  """Parses a QBrush instance."""
  return _createBrush(*args, **kwargs)


def emptyBrush() -> QBrush:
  """Creates an empty QBrush instance."""
  return _createBrush(QColor(0, 0, 0, 0), Qt.BrushStyle.NoBrush, )


def solidBrush(*args, **kwargs) -> QBrush:
  """Creates a solid QBrush instance."""
  color = parseColor(*args, **kwargs)
  return _createBrush(color, Qt.BrushStyle.SolidPattern, )



"""HorizontalLayout wraps the QHBoxLayout."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QHBoxLayout

from vistside.core import parseParent


class HorizontalLayout(QHBoxLayout):
  """HorizontalLayout wraps the QHBoxLayout."""

  def __init__(self, *args, **kwargs) -> None:
    parent = parseParent(*args)
    QHBoxLayout.__init__(self, parent)
    self.setContentsMargins(2, 2, 2, 2, )
    self.setSpacing(2)



"""The core module provides utilities and short names enums of PySide6. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._keyboard_shortcuts import KeyboardShortcuts
from ._parse_parent import parseParent
from ._align_enum import AlignmentEnum

from ._colors import *  # Only acceptable use of wildcard import
from ._qt_enums import *  # Only acceptable use of wildcard import

from ._white_noisinator import generateWhiteNoise, interpolateWhiteNoise
from ._array_field import ArrayField
from ._right_now_field import RightNowField
from ._data_roll import DataRoll, DataRollField
from ._color_factory import parseColor
from ._brush_factory import parseBrush
from ._brush_field import BrushField
from ._pen_factory import parsePen
from ._pen_field import PenField
from ._font_factory import resolveFontFamily, parseFont
from ._font_field import FontField
from ._empty_pen import EmptyPen
from ._empty_brush import EmptyBrush
from ._text_pen import TextPen
from ._font_pen_field import FontPenField
from ._timer_factory import parseTimer
from ._timer_field import TimerField
